package media

import (
	"fmt"
	"image"
	"image/jpeg"
	"io"
	"os"
	"path"
	"path/filepath"
	"regexp"
	"strconv"

	// for processing images
	_ "image/gif"
	_ "image/png"

	"github.com/nfnt/resize"
)

type MediaSource struct {
	StorageDir string
	ThumbDir   string
}

func (ms *MediaSource) LocationOrig(m Media) string {
	return path.Join(ms.StorageDir, m.Name)
}

func (ms *MediaSource) ThumbPath(m Media, size image.Rectangle) string {
	w := size.Dx()
	h := size.Dy()

	var ws, hs string

	if w != 0 {
		ws = fmt.Sprint(w)
	}
	if h != 0 {
		hs = fmt.Sprint(h)
	}

	thumbSlug := filepath.Join(fmt.Sprintf("%sx%s", ws, hs), m.Name)
	return path.Join("/media", thumbSlug)
}

func (ms *MediaSource) ThumbFilename(m Media, size image.Rectangle) string {
	return filepath.Join(ms.ThumbDir, ms.ThumbPath(m, size))
}

func (ms *MediaSource) ReceiveNewMedia(name string, r io.Reader) error {
	dest := path.Join(ms.StorageDir, name)
	f, err := os.OpenFile(dest, os.O_WRONLY|os.O_CREATE, 0644)
	if err != nil {
		return err
	}

	_, err = io.Copy(f, r)
	if err != nil {
		return err
	}

	return f.Close()
}

type Media struct {
	Type string
	Name string
	Size image.Rectangle
}

func (ms *MediaSource) Size(name string) (image.Rectangle, error) {
	f, err := os.Open(name)
	if err != nil {
		return image.ZR, err
	}
	defer f.Close()

	cfg, _, err := image.DecodeConfig(f)
	if err != nil {
		return image.ZR, err
	}

	width := cfg.Width
	height := cfg.Height

	return image.Rect(0, 0, width, height), nil
}

func (ms *MediaSource) ThumbMax(m Media, maxDim int) (string, image.Rectangle, error) {
	f, err := os.Open(ms.LocationOrig(m))
	if err != nil {
		return "", image.ZR, err
	}
	defer f.Close()

	cfg, _, err := image.DecodeConfig(f)
	if err != nil {
		return "", image.ZR, err
	}

	width := cfg.Width
	height := cfg.Height

	if width > height {
		height = height * maxDim / width
		width = maxDim
	} else {
		width = width * maxDim / height
		height = maxDim
	}

	size := image.Rect(0, 0, width, height)
	if ms.HasThumb(m, size) {
		return ms.ThumbPath(m, size), size, nil
	}

	_, err = f.Seek(0, io.SeekStart)
	if err != nil {
		return "", image.ZR, err
	}

	img, _, err := image.Decode(f)
	if err != nil {
		return "", image.ZR, err
	}

	src, err := ms.ThumbImage(img, m, size)
	return src, size, err
}

func (ms *MediaSource) HasThumb(m Media, size image.Rectangle) bool {
	_, err := os.Stat(ms.ThumbFilename(m, size))
	return err == nil
}

func (ms *MediaSource) ByName(name string) *Media {
	size, _ := ms.Size(path.Join(ms.StorageDir, name))
	return &Media{
		Type: "image",
		Name: name,
		Size: size,
	}
}

func (ms *MediaSource) Thumb(m Media, size image.Rectangle) (string, error) {
	if ms.HasThumb(m, size) {
		return ms.ThumbPath(m, size), nil
	}

	f, err := os.Open(ms.LocationOrig(m))
	if err != nil {
		return "", err
	}
	defer f.Close()

	img, _, err := image.Decode(f)
	if err != nil {
		return "", err
	}

	return ms.ThumbImage(img, m, size)
}

func (ms *MediaSource) ThumbImage(img image.Image, m Media, size image.Rectangle) (string, error) {

	thumbLoc := ms.ThumbFilename(m, size)
	os.MkdirAll(path.Dir(thumbLoc), 0755)
	fthumb, err := os.OpenFile(thumbLoc, os.O_CREATE|os.O_WRONLY|os.O_TRUNC, 0655)
	if err != nil {
		return "", err
	}

	img = resize.Resize(uint(size.Dx()), uint(size.Dy()), img, resize.Bilinear)

	err = jpeg.Encode(fthumb, img, nil)
	if err != nil {
		return "", err
	}

	err = fthumb.Close()
	if err != nil {
		return "", err
	}

	return ms.ThumbPath(m, size), nil
}

func (ms *MediaSource) Walk() ([]*Media, error) {
	var media []*Media

	err := filepath.Walk(ms.StorageDir, func(name string, fi os.FileInfo, err error) error {
		if err != nil {
			return err
		}
		if fi.IsDir() {
			return nil
		}
		media = append(media, ms.ByName(path.Base(name)))
		return nil
	})

	return media, err
}

var (
	sizeString = regexp.MustCompile(`([0-9]*)(x)?([0-9]*)`)
)

func ParseSizeString(str string, actual image.Rectangle) (image.Rectangle, error) {
	var err = fmt.Errorf("expected a size string {width}x{height}, saw %q", str)

	strs := sizeString.FindStringSubmatch(str)
	if len(strs) < 4 {
		return image.ZR, err
	}

	var w, h int
	var strconvErr error

	if strs[1] != "" {
		w, strconvErr = strconv.Atoi(strs[1])
		if strconvErr != nil {
			return image.ZR, err
		}
	}

	if strs[3] != "" {
		h, strconvErr = strconv.Atoi(strs[3])
		if strconvErr != nil {
			return image.ZR, err
		}
	}

	if strs[2] != "x" {
		// w was the only dimension given, so set it to the greater dimension
		// of the actual image size
		if actual.Dx() > actual.Dy() {
			h = 0
		} else {
			h = w
			w = 0
		}
	}

	return image.Rect(0, 0, w, h), nil
}

func removeExtension(name string) string {
	ext := path.Ext(name)
	return name[:len(name)-len(ext)]
}
